/* Copyright (C) 2013 MORIOKA Tomohiko
   This file is part of the CONCORD Library.

   The CONCORD Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The CONCORD Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the CONCORD Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <stdlib.h>
#include <ctype.h>
#include "cos-read.h"

COS_object
cos_read_int (unsigned char *str, size_t len, size_t start, size_t* endp)
{
  size_t i = start;
  int c;
  int negative_flag;
  int dest;

  if ( i < len )
    {
      switch ( str[i] )
	{
	case '+':
	  negative_flag = 0;
	  i++;
	  break;
	case '-':
	  negative_flag = 1;
	  i++;
	  break;
	default:
	  negative_flag = 0;
	}

      if ( (i < len) && (c = str[i++])
	   && ('0' <= c) && (c <= '9') )
	{
	  dest = c - '0';

	  while ( i < len )
	    {
	      c = str[i];
	      if ( ('0' <= c) && (c <= '9') )
		{
		  dest = dest * 10 + c - '0';
		  i++;
		}
	      else if ( isspace (c) )
		{
		  *endp = i;
		  return cos_make_int ( negative_flag ? - dest : dest );
		}
	      else
		return NULL;
	    }
	  *endp = i;
	  return cos_make_int ( negative_flag ? - dest : dest );
	}
    }
  return NULL;
}


int
cos_read_utf8 (unsigned char *str, size_t len, size_t start, size_t* endp)
{
  size_t i = start;

  if ( len >= i + 1 )
    {
      unsigned char c = str[i++];
      int counter;
      int cid;

      if ( c < 0xC0 )
	{
	  cid = c;
	  counter = 0;
	}
      else if ( c < 0xE0 )
	{
	  cid = c & 0x1f;
	  counter = 1;
	}
      else if ( c < 0xF0 )
	{
	  cid = c & 0x0f;
	  counter = 2;
	}
      else if ( c < 0xF8 )
	{
	  cid = c & 0x07;
	  counter = 3;
	}
      else if ( c < 0xFC )
	{
	  cid = c & 0x03;
	  counter = 4;
	}
      else
	{
	  cid = c & 0x01;
	  counter = 5;
	}

      if (counter + i <= len)
	{
	  int j;

	  for (j = 0; j < counter; j++)
	    cid = (cid << 6) | (str[i++] & 0x3F);
	  *endp = i;
	  return cid;
	}
    }
  return -1;
}

int
cos_read_char (unsigned char *str, size_t len, size_t start, size_t* endp)
{
  size_t i = start;

  if ( (len >= start + 2) && (str[i++] == '?') )
    {
      if ( (len >= start + 3) && (str[i] == '\\') )
	{
	  i++;
	  return cos_read_utf8 (str, len, i, endp);
	}
      else
	return cos_read_utf8 (str, len, i, endp);
    }

  return -1;
}


COS_String
cos_read_string (unsigned char *str, size_t len, size_t start, size_t* endp)
{
  size_t i = start;
  int c;

  if ( (len >= start + 2) && (str[i++] == '"') )
    {
      while ( ( i < len )
	      && ( (c = cos_read_utf8 (str, len, i, &i)) >= 0 )
	      )
	{
	  if ( c == '"' )
	    {
	      return cos_make_string ((char*)&str[1], i - 2);
	    }
	  else if ( c == '\\' )
	    {
	      i++;
	      if ( cos_read_utf8 (str, len, i, &i) < 0 )
		return NULL;
	    }
	}
    }
  return NULL;
}
